/*
 * Copyright (c) 1999, Boris Popov
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Boris Popov.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ipxping.c,v 1.1.2.1 1999/12/10 09:07:07 bp Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netipx/ipx.h>

#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>
#include <sysexits.h>
#include <signal.h>

#include <netncp/ncp_lib.h>

#ifndef IPXPORT_IPXPING
#define IPXPORT_IPXPING		0x9086
#endif

#define PING_DATA_LEN		8

static char pingdata[PING_DATA_LEN] = "FreeBSD ";
int stop_flag;

static int ipx_recv(int sock, void *buf, int len, int flags, int timeout);
static void usage(void);
static void inthandler(int sig);
static void tvsub(struct timeval *, struct timeval *);

int
main(int argc, char *argv[]) {
	struct ipx ipx;
	struct ipx_addr dest;
	struct sockaddr_ipx sipx;
	struct ncp_conn_loginfo li;
	struct timeval now, last, tv;
	struct sigaction si_sa;
	double rtt, rtmin, rtmax, rtsum;
	int opt, sock, slen, rlen, count, delay, nreceived, nsent, i, error;
	u_short seq;
	u_char rq[10 + PING_DATA_LEN], rp[10 + PING_DATA_LEN], *p;
	char *ep;

	delay = 1;
	count = -1;
	seq = 0;
	rtmin = 999999999.0;
	rtmax = 0.0;
	rtsum = 0.0;

	while ((opt = getopt(argc, argv, "c:i:")) != EOF) {
		switch (opt) {
		    case 'c':
			i = strtoul(optarg, &ep, 0);
			if (*ep || ep == optarg || i > LONG_MAX || !i)
				errx(EX_USAGE,
				    "invalid count of packets to transmit: `%s'",
				    optarg);
			count = i;
			break;
		    case 'i':
			i = strtoul(optarg, &ep, 0);
			if (*ep || ep == optarg || i > LONG_MAX || !i)
				errx(EX_USAGE,
				    "invalid delay interval: `%s'",
				    optarg);
			delay = i;
			break;
		    default:
			usage();
			/* NOTREACHED */
		}
	}

	if (argc - optind != 1)
		usage();
	ep = argv[optind];
	if (isdigit(ep[0])) {
		dest = ipx_addr(ep);
	} else {
		if (ncp_initlib() != 0)
			exit(1);
		ncp_li_init(&li, 0, NULL);
		ncp_li_setserver(&li, ep);
		error = ncp_find_fileserver(&li, AF_IPX, NULL);
		if (error) {
			ncp_error("could not find server %s\n", error, li.server);
			exit(1);
		}
		dest = li.ipxaddr.sipx_addr;
		ncp_li_done(&li);
	}
	if (ipx_nullnet(dest) || ipx_nullhost(dest))
		errx(EX_DATAERR, "invalid address");

	sigemptyset(&si_sa.sa_mask);
	si_sa.sa_flags = 0;
	si_sa.sa_handler = inthandler;
	if (sigaction(SIGINT, &si_sa, 0) == -1) {
	        err(EX_OSERR, "sigaction SIGINT");
						        }
	sock = socket(AF_IPX, SOCK_DGRAM, 0);
	if (sock < 0) {
		errx(1, "socket()");
	}
	slen = sizeof(sipx);
	bzero(&sipx, slen);
	sipx.sipx_family = AF_IPX;
	sipx.sipx_len = slen;
	if (bind(sock, (struct sockaddr*)&sipx, slen) == -1) {
		close(sock);
		errx(1, "bind()");
	}

	ipx.ipx_pt = IPXPROTO_PXP;
	if (setsockopt(sock, 0, SO_DEFAULT_HEADERS, &ipx, sizeof(ipx)))
		errx(1, "setsockopt()");

	bzero(&sipx, slen);
	sipx.sipx_family = AF_IPX;
	sipx.sipx_len = slen;
	sipx.sipx_addr = dest;
	sipx.sipx_addr.x_port = htons(IPXPORT_IPXPING);
	printf("Ping %s (%s): %d data bytes\n",
	    ep, ipx_ntoa(sipx.sipx_addr), PING_DATA_LEN);

	bcopy("Ping", rq, 4);
	p = rq + 4;
	*p++ = 1;	/* version */
	*p++ = 0;	/* request */
	p += 2;
	*p++ = 0;	/* result */
	*p++ = 0;	/* reserved */
	bcopy(pingdata, p, PING_DATA_LEN);
	nreceived = 0;
	nsent = 0;
	do {
		*(u_short*)(rq + 6) = seq;
		(void)gettimeofday(&last, NULL);
		rlen = sendto(sock, rq, sizeof(rq), 0,
			(struct sockaddr*)&sipx, slen);
		if (rlen < 0) {
			warn("Can't sendto");
		} else
			nsent++;

		rlen = ipx_recv(sock, rp, sizeof(rp), 0, delay);
		(void)gettimeofday(&now, NULL);
		if (rlen == -1) {
			warn("can't recv()");
		} else if (rlen < sizeof(rp)) {
			warn("reply too short");
		} else if (memcmp("Ping", rp , 4) != 0) {
			warnx("signature mismatch");
		} else if (memcmp(pingdata, rp + 10, PING_DATA_LEN) != 0) {
			warnx("reply data mismatch");
		} else if (*(u_short*)(rp + 6) != seq) {
			warnx("reply sequence mismatch: %d:%d", 
			    *(u_short*)(rp + 6), seq);
		} else {
			nreceived++;
			memcpy(&tv, &now, sizeof(tv));
			tvsub(&tv, &last);
			rtt = ((double)tv.tv_sec) * 1000.0 +
                              ((double)tv.tv_usec) / 1000.0;
			if (rtt < rtmin)
				rtmin = rtt;
			if (rtt > rtmax)
				rtmax = rtt;
			rtsum += rtt;
			printf("%d bytes received: seq=%d time=%.3f ms\n",
				rlen - 10, seq, rtt);
		}
		if (stop_flag)
			break;
		if (count != -1) {
			count--;
			if (count == 0)
				break;
		}
		seq++;
		i = delay - (now.tv_sec - last.tv_sec);
		if (i > 0)
			sleep(i);
	} while(1);

	printf("\n--- %s ping statistics ---\n", ep);
	printf("%d packets sent, %d received", nsent, nreceived);

	if (nsent) {
		printf(", %d%% packets loss",
		       ((nsent - nreceived) * 100) / nsent);
	}
	printf("\n");
	if (nreceived) {
		printf("round-trip min/avg/max = %.3f/%.3f/%.3f\n",
			rtmin, rtsum / nsent, rtmax);
	}
	exit(nreceived ? 0 : 1);
}

static int
ipx_recv(int sock, void *buf, int len, int flags, int timeout) {
	fd_set rd, wr, ex;
	struct timeval tv;
	int result;

	FD_ZERO(&rd);
	FD_ZERO(&wr);
	FD_ZERO(&ex);
	FD_SET(sock, &rd);

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	if ((result = select(sock + 1, &rd, &wr, &ex, &tv)) == -1) {
		return result;
	}
	if (FD_ISSET(sock, &rd)) {
		result = recv(sock, buf, len, flags);
	} else {
		errno = ETIMEDOUT;
		result = -1;
	}
	return result;
}

void
inthandler(int sig)
{
	stop_flag = 1;
}

static void
tvsub(struct timeval *out, struct timeval *in)
{
	if ((out->tv_usec -= in->tv_usec) < 0) {
		--out->tv_sec;
		out->tv_usec += 1000000;
	}
	out->tv_sec -= in->tv_sec;
}

static void
usage(void) {
	fprintf(stderr, "%s\n",
	    "usage: ipxping [-c count] [-i delay] [network.]host"
	);
	exit(EX_USAGE);
}
