/*
 * Netware print server daemon
 *
 * Copyright (C) 1996 by Volker Lendecke
 * Copyright (c) 1999, Boris Popov
 *
 * $Id: ncplpd.c,v 1.4.2.1 1999/09/30 10:31:01 bp Exp $
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <syslog.h>
#include <fcntl.h>
#include <signal.h>

#include <netncp/ncp_lib.h>
#include <netncp/ncp_rcfile.h>

struct nw_queue {
	int	connid;
	char	queue_name[NCP_BINDERY_NAME_LEN];
	u_int32_t queue_id;
	u_int16_t job_type;
	char *command;
};

static struct nw_queue q;

static int term_request;
extern char *__progname;

static int init_queue(int connid, char *queue_name, char *command, struct nw_queue *q);
static int poll_queue(struct nw_queue *q);

static void
usage(void) {
	fprintf(stderr, "usage: %s [options] file\n", __progname);
	exit(1);
}

static void
help(void) {
	printf("\n");
	printf("usage: %s [options]\n", __progname);
	printf("\n"
	       "-S server     	Server name to be used\n"
	       "-U username	Print Server name sent to server\n"
	       "[....]		other conn options\n"
	       "-q queue name	Name of the printing queue to use\n"
	       "-c command	Name of print command, default: 'lpr'\n"
	       "-j job type	Type of job (Form number) to service\n"
	       "-t timeout	Polling interval, default: 30 sec\n"
	       "-d		Debug: don't daemonize\n"
	       "-h		print this help text\n"
	       "\n");
}

static void
terminate_handler(int dummy) {
	signal(SIGTERM, terminate_handler);
	signal(SIGINT, terminate_handler);
	term_request = 1;
}

int
main(int argc, char *argv[]) {
	int connid, poll_timeout, opt, job_type = 0xffff;
	int debug = 0, i, error;
	char *queue_name = NULL;
	char default_command[] = "lpr";
	char *command = default_command;
	struct ncp_conn_loginfo li;

	poll_timeout = 10;
	for (i = 1; i < argc; i += 1) {
		if (strcmp(argv[i], "-h") == 0) {
			help();
			exit(0);
		}
	}
	if(ncp_initlib()) exit(1);

	if (ncp_li_init(&li,argc, argv)) return 1; /* this should help to share conn */
	if (ncp_li_readrc(&li)) return 1;
	if (ncp_rc) {
		/* XXX read defaults for queue ??? */
		rc_close(ncp_rc);
	}
	while ((opt = getopt(argc, argv, STDPARAM_OPT"q:c:j:t:dh")) != EOF) {
		switch (opt) {
		    case STDPARAM_ARGS:
			if (ncp_li_arg(&li, opt, optarg)) {	
				return 1;
			}
			break;
		    case 'q':
			queue_name = optarg;
			break;
		    case 'c':
			command = optarg;
			break;
		    case 'j':
			job_type = atoi(optarg);
			break;
		    case 't':
			poll_timeout = atoi(optarg);
			break;
		    case 'd':
			debug = 1;
			break;
		    case 'h':
			break;
		    default:
			printf("Invalid option\n");
			usage();
			return -1;
		}
	}

	if (argc != optind) {
		usage();
		return -1;
	}
	if (queue_name == NULL)	{
		fprintf(stderr, "You must specify a queue to attach\n");
		return 1;
	}
	bzero(&q,sizeof(q));
	if (ncp_li_check(&li)) return 1;
	/*
	 * must become a daemon before loggin in, or connection will belong
	 * to parent
	 */
	if (debug == 0)	{
		daemon(0,0);
		openlog("ncplpd", LOG_PID, LOG_LPR);
		syslog(LOG_INFO,"started");
	}
	/* now we can try to login, or use already established connection */
	li.objtype = NCP_BINDERY_PSERVER;
	error = ncp_li_login(&li,&connid);
	if (error) {
		syslog(LOG_ERR,"Cannot login to server %s,%s\n",li.server,strerror(errno));
		ncp_disconnect(connid);
		exit(1);
	}

	if (init_queue(connid, queue_name, command, &q) != 0) {
		syslog(LOG_ERR,"Could not init queue %s",queue_name);
		ncp_disconnect(connid);
		return 1;
	}
	q.job_type = job_type;

	term_request = 0;
	signal(SIGTERM, terminate_handler);
	signal(SIGINT, terminate_handler);

	while (1) {
		if (debug) {
			fprintf(stderr, "Polling ...\n");
		}
		if ((poll_queue(&q) != 0) && (term_request == 0)) {
			continue;
		}
		if (term_request != 0) {
			break;
		}
		sleep(poll_timeout);
	}
	syslog(LOG_ERR,"Catch a signal, terminate.\n");
	if (debug)
		fprintf(stderr,"Catch a signal, terminate.\n");
	ncp_detach_from_queue(connid, q.queue_id);

	ncp_disconnect(connid);
	return 0;
}

static int
init_queue(int connid, char *queue_name, char *command, struct nw_queue *q) {
	struct ncp_bindery_object obj;

	ncp_str_upper(queue_name);

	q->connid = connid;
	q->command = command;
	if (ncp_get_bindery_object_id(connid, NCP_BINDERY_PQUEUE,
				      queue_name, &obj) != 0) {
		syslog(LOG_ERR, "Queue %s not found\n", queue_name);
		return -1;
	}
	q->queue_id = obj.object_id;
	memcpy(q->queue_name, obj.object_name, sizeof(q->queue_name));

	if (ncp_attach_to_queue(connid, q->queue_id) != 0) {
		syslog(LOG_ERR,"Could not attach to queue %s\n", queue_name);
		return -1;
	}
	return 0;
}


void
build_command(struct nw_queue *q, struct queue_job *j, char *target, int target_size, char *user)
{
	char *s = q->command;
	char *target_end = target + target_size;

	void add_string(const char *s) {
		int len = strlen(s);
		if (target + len + 1 > target_end)
		{
			len = target_end - target - 1;
		}
		strncpy(target, s, len);
		target += len;
	}

	bzero(target, target_size);

	while ((*s != 0) && (target < target_end)) {
		if (*s != '%') {
			*target = *s;
			target += 1;
			s += 1;
			continue;
		}
		switch (*(s + 1)) {
		case '%':
			*target = '%';
			target += 1;
			break;
		case 'u':
		        add_string(user);
			break;
                case 'd':
                        if (j->j.JobTextDescription[0])
                                add_string(j->j.JobTextDescription);
                        else
                                add_string("No Description");
                        break;
		default:
			*target = '%';
			*(target + 1) = *(s + 1);
			target += 2;
		}
		s += 2;
	}
}

static int
poll_queue(struct nw_queue *q) {
	struct queue_job job;
	int fd[2];
	int pid;
	int retcode;
	struct ncp_bindery_object u;
	char user[50];

	if (ncp_service_queue_job(q->connid, q->queue_id, q->job_type,
				  &job) != 0) {
		/* No job for us */
		return 0;
	}

        if (ncp_get_queue_job_info(q->connid, q->queue_id, job.j.JobNumber, &job.j) != 0)
        {
                job.j.JobTextDescription[0]=0;
        }

	if ((retcode=ncp_get_bindery_object_name
	     (q->connid, htonl(job.j.ClientObjectID), &u)) == 0) {
	    memcpy(user,u.object_name,48);
            user[48]=0;
	} else {
	    sprintf(user,"<Unknown>");
	}
	
	if (pipe(fd) < 0) {
		syslog(LOG_ERR, "pipe error: %m");
		goto fail;
	}
	if ((pid = fork()) < 0)	{
		syslog(LOG_ERR, "fork error: %m");
		goto fail;
	}
	if (pid > 0) {
		/* parent */
		char buf[1024];
		size_t result;
		off_t offset = 0;

		close(fd[0]);	/* close read end */

		while ((result = ncp_read(q->connid, &job.file_handle, offset,
					  sizeof(buf), buf)) > 0)
		{
			offset += result;
			if (write(fd[1], buf, result) != result) {
				goto fail;
			}
		}

		close(fd[1]);	/* and close write end */

		if (waitpid(pid, NULL, 0) < 0) {
			syslog(LOG_ERR, "waitpid: %m\n");
		}
	} else {
		/* child */
		char command[2048];

		close(fd[1]);	/* close write end */
		if (fd[0] != STDIN_FILENO) {
			if (dup2(fd[0], STDIN_FILENO) != STDIN_FILENO) {
				syslog(LOG_ERR, "dup2 error: %m\n");
				close(fd[0]);
				exit(1);
			}
			close(fd[0]);
		}
		build_command(q, &job, command, sizeof(command), user);
		execl("/bin/sh", "sh", "-c", command, NULL);
		syslog(LOG_ERR, "exec error: %m\n");
		close(fd[0]);
		exit(1);
	}
	ncp_finish_servicing_job(q->connid, q->queue_id, job.j.JobNumber, 0);
	return 1;

fail:
	ncp_abort_servicing_job(q->connid, q->queue_id, job.j.JobNumber);
	/* We tell that we did not have a job to avoid overloading
	   when something's wrong */
	return 0;
}
