/*
 * Put job to a NetWare print queue.
 *
 * Copyright (C) 1995 by Volker Lendecke
 * Copyright (C) 1999 by Boris Popov
 *
 * $Id: ncprint.c,v 1.8.2.1 1999/09/30 10:31:02 bp Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include <netncp/ncp_lib.h>
#include <netncp/ncp_rcfile.h>

/*
 * These are declared in global space to reduce number of parms
 */
static struct queue_job j;
static struct print_job_record pj;
static int banner_flag = PRINT_BANNER;

extern char *__progname;

struct ncp_args ncprint_args[] = {
	{NCA_STR,'p', "path_name",NAFL_NONE,sizeof (pj.Path)-1, ""},
	{NCA_STR,'b', "banner_name",NAFL_NONE,sizeof (pj.BannerName)-1, ""},
	{NCA_STR,'d', "job_desc",NAFL_NONE,sizeof (j.j.JobTextDescription)-1, ""},
	{NCA_STR,'f', "file_name",NAFL_NONE,sizeof (pj.FnameBanner)-1, ""},
	{NCA_INT,'l', "lines",	NAFL_MINMAX,66,"",0,65535},
	{NCA_INT,'r', "rows",	NAFL_MINMAX,80,"",0,65535},
	{NCA_INT,'c', "copies",	NAFL_MINMAX, 1,"",0,65000},
	{NCA_INT,'t', "tab_size",NAFL_MINMAX,8,"",0,255},
	{NCA_BOOL,'n',"no_form_feed",NAFL_NONE,1},
	{NCA_BOOL,'s',"no_banner",NAFL_NONE,1},
	{NCA_INT,'j', "form_number",NAFL_MINMAX,0,"",0,65535},
	{NCA_INT,0,""}
};

static void usage(void);
static int read_rc_file(struct ncp_conn_loginfo *li,char **queue);
static int my_setopt(struct ncp_args *na);

int
my_setopt(struct ncp_args *na) {
	switch(na->opt) {
	    case 'p': /* Path */
		pj.CtrlFlags |= banner_flag;
		strncpy(pj.Path, na->str, na->ival);
		break;
	    case 'b': /* Banner Name */
		pj.CtrlFlags |= banner_flag;
		strncpy(pj.BannerName, na->str, na->ival);
		break;
	    case 'f': /* File Name in Banner */
		pj.CtrlFlags |= banner_flag;
		strncpy(pj.FnameBanner, na->str, na->ival);
		break;
	    case 'd': /* Job Description */
		pj.CtrlFlags |= banner_flag;
		strncpy(j.j.JobTextDescription, na->str, na->ival);
		break;
	    case 'l': /* lines, default: 66 */
		pj.Lines = htons(na->ival);
		pj.CtrlFlags |= EXPAND_TABS;
		break;
	    case 'r': /* rows, default: 80 */
		pj.Rows = htons(na->ival);
		pj.CtrlFlags |= EXPAND_TABS;
		break;
	    case 'c': /* copies, default: 1 */
		pj.Copies = htons(na->ival);
		pj.CtrlFlags |= EXPAND_TABS;
		break;
	    case 't': /* tab size, default: 8 */
		pj.TabSize = na->ival;
		pj.CtrlFlags |= EXPAND_TABS;
		break;
	    case 'n': /* no form feed */
		pj.CtrlFlags |= NO_FORM_FEED;
		break;
	    case 's': /* no banner */
		banner_flag = 0;
		pj.CtrlFlags &= ~PRINT_BANNER;
		break;
	    case 'j': /* Form number, default: 0 */
		j.j.JobType = htons(na->ival);
		break;
	}
	return 0;
}

int 
read_rc_file(struct ncp_conn_loginfo *li,char **queue) {
	static char rc_queue[NCP_BINDERY_NAME_LEN*2];
	char uname[NCP_BINDERY_NAME_LEN*2+1], *sect = uname;

	strcat(strcat(strcpy(uname,li->server),":"),li->user);
	if (rc_getstring(ncp_rc,sect,"default_queue",sizeof(rc_queue),rc_queue) == 0)
		*queue = rc_queue;
	strcat(strcat(strcpy(uname,li->server),":"),*queue);
	return ncp_args_parserc(ncprint_args, sect, my_setopt);
}

int
main(int argc, char *argv[]) {
	char *queue = NULL;
	int opt, i, connid, file, error;
	int offset, count;
	char buf[8192];
	char *file_name;
	struct ncp_conn_loginfo li;
	struct ncp_bindery_object q;

	if (argc == 1 || (argc == 2 && strcmp(argv[1], "-h") == 0)) {
		usage();
		exit(0);
	}
	if(ncp_initlib()) exit(1);
	bzero(&j, sizeof(j));
	bzero(&pj, sizeof(pj));
	bzero(&q, sizeof(q));

	/*
	 * Fill in default values for print job
	 */
	j.j.TargetServerID = 0xffffffff;	/* any server */
	/* at once */
	memset(&(j.j.TargetExecTime), 0xff, sizeof(j.j.TargetExecTime));
	j.j.JobType = htons(0);
	strcpy(j.j.JobTextDescription, "No Description");

	pj.Version = 0;
	pj.TabSize = 8;
	pj.Copies = htons(1);
	pj.CtrlFlags = 0;
	pj.Lines = htons(66);
	pj.Rows = htons(80);
	strcpy(pj.FnameHeader, "stdin");
	/*
	 * Check for queue name either in rc file or cmd line
	 * because we will read rc file before main parse of cmd line
	 */
	if (ncp_li_init(&li, argc, argv)) return 1;
	if (ncp_li_readrc(&li)) return 1;
	while ((opt = ncp_getopt(argc, argv, ":q:")) != -1) {
		switch (opt) {
		    case 'q':
			if (strlen(ncp_optarg) >= NCP_BINDERY_NAME_LEN) {
				fprintf(stderr,"queue name too long: %s\n", ncp_optarg);
				exit(1);
			}
			queue = ncp_optarg;
			break;
		}
	}
	if (queue == NULL) {
		fprintf(stderr,"queue name should be specified\n");
		return 1;
	}
	ncp_optind = ncp_optreset = 1;
	ncp_str_upper(queue);
	if (ncp_rc) {
		error = read_rc_file(&li,&queue);
		rc_close(ncp_rc);
		if (error) return 1;
	}
	while ((opt = getopt(argc, argv, STDPARAM_OPT"hq:d:p:b:f:l:r:c:t:j:ns")) != EOF) {
		switch (opt) {
		    case STDPARAM_ARGS:
			if (ncp_li_arg(&li, opt, optarg)) {	
				return 1;
			}
			break;
		case 'h': case '?':
			usage();
			exit(1);
		case 'p':
		case 'b': /* Banner Name */
		case 'f': /* File Name in Banner */
		case 'l': /* lines, default: 66 */
		case 'r': /* rows, default: 80 */
		case 'c': /* copies, default: 1 */
		case 't': /* tab size, default: 8 */
		case 'n': /* no form feed */
		case 's': /* no banner */
		case 'j': /* Form number, default: 0 */
		case 'd': /* Job Description */
			error = ncp_args_parseopt(ncprint_args,opt,optarg,my_setopt);
			if (error) return 1;
			break;
		case 'q': break;
		default:
			usage();
			exit(1);
		}
	}

	if (optind != argc - 1)	{
		printf("No file name specified.\n");
		usage();
		exit(1);
	}
	file_name = argv[optind];

	if (strcmp(file_name, "-") == 0) {
		file = 0;	/* stdin */
	} else {
		file = open(file_name, O_RDONLY, 0);
		if (file < 0) {
			printf("could not open file %s:\n",file_name);
			perror("");
			exit(1);
		}
		strncpy(pj.FnameHeader, file_name, sizeof(pj.FnameHeader)-1);

		if (strlen(pj.FnameBanner) == 0) {
			strncpy(pj.FnameBanner, file_name, sizeof(pj.FnameBanner)-1);
		}
	}

	memcpy(j.j.ClientRecordArea, &pj, sizeof(pj));

	error = ncp_li_check(&li);
	if (error) exit(1);
	error = ncp_li_login(&li,&connid);
	if( error ){
		fprintf(stderr,"Cannot login to server %s,%s\n",li.server,strerror(errno));
		ncp_disconnect(connid);
		exit(1);
	}
	if (ncp_scan_bindery_object(connid, 0xffffffff, NCP_BINDERY_PQUEUE,
				    queue, &q) != 0) {
		fprintf(stderr,"could not find queue %s\n", queue);
		ncp_disconnect(connid);
		exit(1);
	}
	if (ncp_create_queue_job_and_file(connid, q.object_id, &j) != 0) {
		printf("can't create queue job\n");
		ncp_disconnect(connid);
		exit(1);
	}
	offset = 0;
	error = 0;
	for(;;) {
		count = read(file, buf, sizeof(buf));
		if (count == 0) break;
		if (count < 0) {
			fprintf(stderr, "error reading input file %s\n",strerror(errno));
			error = errno;
			break;
		}
		i = ncp_write(connid, &j.file_handle, offset, count, buf);
		if (i < 0) {
			fprintf(stderr, "error writing job %s\n",strerror(errno));
			error = errno;
			break;
		}
		if ( i != count) {
			fprintf(stderr, "incomplete write to job\n");
			error = EIO;
			break;
		}
		offset += count;
	}
	close(file);
	if (error) {
		ncp_disconnect(connid);
		return (1);
	}
	if (ncp_close_file_and_start_job(connid, q.object_id, &j) != 0) {
		fprintf(stderr,"can't start printing\n");
	}
	ncp_disconnect(connid);
	return 0;
}


static void
usage(void) {
	printf("\n");
	printf("usage: %s [options] file\n", __progname);
	printf("\n"
	       "-S server      Server name to be used\n"
	       "-U username    Username sent to server\n"
	       "[other conn params ...]\n"
	       "-q queue name  Name of the printing queue to use\n"
	       "-b bannername  Banner name (up to 12 chars)\n"
	       "-c copies      Number of copies (1)\n"
	       "-d job desc    Job description\n"
	       "-f filename    Filename to appear on banner\n"
	       "-h             print this help text\n"
	       "-j form #      Form number to print on\n"
	       "-l lines       Number of lines per page\n"
	       "-n             Suppress print server form feeds\n"
	       "-p path name   Path name to appear on banner\n"
	       "-r rows        Number of rows per page\n"
	       "-s             Suppress banner\n"
	       "-t tab         Number of spaces per tab\n"
	       "\n");
}
