/*
 * Copyright (c) 1999, Boris Popov
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Boris Popov.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
	"$Id: ncpsend.c,v 1.2.2.4 1999/10/21 08:44:59 bp Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sysexits.h>

#include <netncp/ncp_lib.h>
#include <netncp/ncp_rcfile.h>

#define MAX_CONNS_PER_USER	256

enum send_action {
	SA_NONE, SA_TO_USER, SA_TO_CONSOLE, SA_POLL, SA_ENABLE, SA_DISABLE
};


static void
usage(void) {
	printf("usage: ncpsend [options] user \"message\"\n");
	printf("       ncpsend [options] @group \"message\"\n");
	printf("       ncpsend [options] -c \"message\"\n");
	printf("       ncpsend [options] -e | -d\n");
	printf("       ncpsend [options] -p\n");
	exit(1);
}

int
composeMsg(char *dest, int maxlen, char *user, char *src) {
	uid_t uid;
	struct passwd *upw;
	char *uname;
	char tname[20];

	uid = getuid();
	upw = getpwuid(uid);
	if (upw) {
		uname = upw->pw_name;
	} else {
		sprintf(tname, "%d", uid);
		uname = tname;
	}
	snprintf(dest, maxlen, "From %s(%s): %s", user, uname, src);
	return 0;
}

char *
errReason(int error) {
	switch(error) {
	    case 1:
		return "bad station number";
	    case 2:
		return "client not logged in";
	    case 3:
		return "client not accepting messages";
	    case 4:
		return "client already has a message";
	    case 150:
		return "no space to store message";
	    default:
		return "unknown";
	}
}

static int
sendtouser(NWCONN_HANDLE connHandle, char *user, char *message) {
	int error, i;
	nuint16 connCnt;
	nuint16	connlist[MAX_CONNS_PER_USER];
	nuint8 reslist[MAX_CONNS_PER_USER];

	if (isdigit(user[0])) {
		connlist[0] = atoi(user);
		connCnt = 1;
	} else {
		error = NWGetObjectConnectionNumbers(connHandle, user, 
		    NCP_BINDERY_USER, &connCnt, connlist, MAX_CONNS_PER_USER);
		if (error) {
			ncp_error("can't get connections list for user %s",
			    error, user);
			return error;
		}
		if (connCnt == 0) {
			ncp_error("no connection found for %s", 0, user);
			return ENXIO;
		}
	}
	error = NWSendBroadcastMessage(connHandle, message, connCnt, 
	    connlist, reslist);
	if (error) {
		ncp_error("can't send message to %s", error, user);
		return error;
	}
	for (i = 0; i < connCnt; i++) {
		if (reslist[i] == 0) continue;
		printf("Message was not sent to connection %d: %s\n", connlist[i], errReason(reslist[i]));
	}
	return 0;
}

static int
sendtogroup(NWCONN_HANDLE connHandle, char *group, char *message) {
	int segid, i, error;
	nuint32 objid;
	struct ncp_bindery_object obj;
	struct nw_property propval;

	for (segid = 1; ; segid++) {
		error = ncp_read_property_value(connHandle, NCP_BINDERY_UGROUP,
		    group, segid, "GROUP_MEMBERS", &propval);
		if (error)
			break;
		for (i = 0; i < sizeof(propval.value); i += sizeof(nuint32)) {
			objid = getdbe(propval.value, i);
			if (objid == 0)
				break;
			error = ncp_get_bindery_object_name(connHandle, objid,
			     &obj);
			if (error)
				continue;
			printf("Sending message to %s\n", obj.object_name);
			sendtouser(connHandle, obj.object_name, message);
		}
		error = 0;
		if (!propval.more_flag)
			break;
	}
	if (error && segid == 1)
		ncp_error("invalid group %s", error, group);
	return error;
}

int
main(int argc, char **argv) {
	NWCONN_HANDLE connHandle;
	nstr8 message[400];
	int error, opt, argsleft = 2;
	char *user = NULL;
	enum send_action action = SA_TO_USER;
	struct ncp_conn_loginfo li;

	if (ncp_initlib()) exit(1);
	if (ncp_li_init(&li, argc, argv)) return 1;
	if (ncp_li_readrc(&li)) return 1;
	if (ncp_rc)
		rc_close(ncp_rc);
	while ((opt = getopt(argc, argv, STDPARAM_OPT"cdehp")) != EOF){
		if (action != SA_TO_USER)
			errx(EX_USAGE, "only one option allowed");
		switch(opt){
		    case STDPARAM_ARGS:
			if (ncp_li_arg(&li, opt, optarg))
				return 1;
			break;
		    case 'c':
			action = SA_TO_CONSOLE;
			argsleft = 1;
			break;
		    case 'd':
			action = SA_DISABLE;
			argsleft = 0;
			break;
		    case 'e':
			action = SA_ENABLE;
			argsleft = 0;
			break;
		    case 'p':
			action = SA_POLL;
			argsleft = 0;
			break;
		}
	}
	if (argsleft) {
		if ((argc - optind) < argsleft )
			usage();
	}
	if (ncp_li_check(&li))
		return 1;
	li.opt |= NCP_OPT_WDOG;

	error = ncp_li_login(&li, &connHandle);
	if (error)
		errx(EX_OSERR, "can't find suitable connection\n");

	switch (action) {
	    case SA_POLL:
		while(NWGetBroadcastMessage(connHandle, message) == 0) {
			printf(">> %s\n", message);
		}
		break;
	    case SA_TO_CONSOLE:
		composeMsg(message, sizeof(message), li.user, argv[optind]);
		error = NWBroadcastToConsole(connHandle, message);
		if (error)
			ncp_error("Can't send message to console", error);
		break;
	    case SA_ENABLE:
		error = NWEnableBroadcasts(connHandle);
		if (error) {
			ncp_error("Can't enable broadcasts", error);
		} else
			printf("Broadcasts enabled\n");
		break;
	    case SA_DISABLE:
		error = NWDisableBroadcasts(connHandle);
		if (error) {
			ncp_error("Can't disable broadcasts", error);
		} else
			printf("Broadcasts disabled\n");
		break;
	    default:
		break;
	}
	if (action != SA_TO_USER) {
		ncp_disconnect(connHandle);
		exit(0);
	}
	user = argv[optind++];
	composeMsg(message, sizeof(message), li.user, argv[optind]);
	if (user[0] == '@') {
		sendtogroup(connHandle, user + 1, message);
	} else
		sendtouser(connHandle, user, message);
	ncp_disconnect(connHandle);
	return 0;
}

