/*
 *  ncpsvc.c
 *
 *  File server remote control
 *
 *  Copyright (c) 1999 by Boris Popov
 *  The idea taken from ncpfs
 *
 * $Id: ncpsvc.c,v 1.1.2.3 1999/09/30 10:31:02 bp Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <strings.h>
#include <time.h>
#include <unistd.h>
#include <locale.h>
#include <math.h>

#include <netncp/ncp_lib.h>
#include <netncp/ncp_rcfile.h>

struct sv_cmd {
	char	*name;
	int	(*proc)(NWCONN_HANDLE cH, int argc, char *argv[]);
	struct sv_cmd *next;
};

static int fs_gettime(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_sync(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_exec(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_down(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_openbindery(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_closebindery(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_enablelogin(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_disablelogin(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_enabletts(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_disabletts(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_loadnlm(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_unloadnlm(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_mount(NWCONN_HANDLE cH, int argc, char *argv[]);
static int fs_dismount(NWCONN_HANDLE cH, int argc, char *argv[]);

extern char *__progname;
static struct ncp_conn_stat cstat;

static void help(void);
static void rqpermanent(void);

static struct sv_cmd svcmdopen[] = {
	{"bindery",	fs_openbindery,	NULL},
	{NULL,		NULL,	0}
};

static struct sv_cmd svcmdclose[] = {
	{"bindery",	fs_closebindery,NULL},
	{NULL,		NULL,	0}
};

static struct sv_cmd svcmdenable[] = {
	{"login",	fs_enablelogin,	NULL},
	{"tts",		fs_enabletts,	NULL},
	{NULL,		NULL,	0}
};

static struct sv_cmd svcmdisable[] = {
	{"login",	fs_disablelogin,NULL},
	{"tts",		fs_disabletts,	NULL},
	{NULL,		NULL,	0}
};

static struct sv_cmd svcmd[] = {
	{"time",	fs_gettime,	NULL},
	{"sync",	fs_sync,	NULL},
	{"exec",	fs_exec,	NULL},
	{ "load",	fs_loadnlm,	NULL},
	{ "unload",	fs_unloadnlm,	NULL},
	{ "mount",	fs_mount,	NULL},
	{ "dismount",	fs_dismount,	NULL},
	{ "down",	fs_down,	NULL},
	{ "open",	NULL,		svcmdopen},
	{ "close",	NULL,		svcmdclose},
	{ "enable",	NULL,		svcmdenable},
	{ "disable",	NULL,		svcmdisable},
/*	{ "set", cmd_set,		NULL},*/
	{NULL,		NULL,	0 }
};

int
main(int argc, char **argv) {
	int error, opt;
	struct ncp_conn_args li;
	char *cmd;
	struct sv_cmd *p;
	NWCONN_HANDLE cH;

	if (ncp_initlib()) exit(1);
	if (ncp_li_init(&li, argc, argv)) return 1;
	if (ncp_li_readrc(&li)) return 1;
	if (ncp_rc)
		rc_close(ncp_rc);

	while ((opt = getopt(argc, argv, STDPARAM_OPT"h")) != EOF) {
		switch (opt) {
		    case STDPARAM_ARGS:
			if (ncp_li_arg(&li, opt, optarg)) {	
				return 1;
			}
			break;
		    case 'h':
			help();
			exit(1);
		    case '?':
			exit(1);
		    default:
			break;
		}
	}
	if (ncp_li_check(&li)) return 1;
	li.opt |= NCP_OPT_WDOG;
	error = ncp_li_login(&li,&cH);
	if (error){
		fprintf(stderr,"Cannot login to server %s,%s\n",li.server,strerror(errno));
		return(1);
	}
	error = ncp_conn_getinfo(cH, &cstat);
	if (error){
		fprintf(stderr,"Can't stat connection\n");
		return(1);
	}
	if (optind >= argc) {
		fprintf(stderr, "%s: Command missing\n", __progname);
		exit(1);
	}
	cmd = argv[optind++];
	for (p = svcmd; p->name;) {
		if (strcmp(p->name, cmd) != 0) {
			p++;
			continue;
		}
		if (p->next != NULL) {
			p = p->next;
			if (optind == argc) {
				fprintf(stderr, "Unexpected end of command\n");
				exit(1);
			}
			cmd = argv[optind++];
			continue;
		}
		if (p->proc == NULL) {
			fprintf(stderr, "%s: Command not implemented\n", __progname);
			return 0;
		}
		p->proc(cH, argc, argv);
		ncp_disconnect(cH);
		return 0;
	}
	fprintf(stderr, "%s: Invalid command: %s\n", __progname, cmd);
	ncp_disconnect(cH);
	return 1;
}

static void
rqpermanent(void) {
	if ((cstat.flags & NCPFL_PERMANENT) || cstat.ref_cnt > 1) return;
	fprintf(stderr, "This command requires permanent connection\n");
	exit(1);
}

static int
fs_gettime(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error, l;
	time_t time;
	char timebuf[40];

	error = ncp_GetFileServerDateAndTime(cH, &time);
	if (error) {
		fprintf(stderr, "Can't get file server time: %04x\n", error);
		return error;
	}
	l = strlen(ctime_r(&time, timebuf));
	if (l) timebuf[l-1] = 0;
	printf("Server time: %s\n", timebuf);
	return 0;
}

static int
fs_sync(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error, opt, maxdiff = 5, doset = 0;
	time_t svt, loct;
	double diff;

	while ((opt = getopt(argc, argv, "d:e")) != EOF) {
		switch (opt) {
		    case 'd':
			maxdiff = atoi(optarg);
			break;
		    case 'e':
			doset = 1;
			break;
		    default:
			return 1;
		}
	}
	error = ncp_GetFileServerDateAndTime(cH, &svt);
	if (error) {
		fprintf(stderr, "Can't get file server time: %04x\n", error);
		return error;
	}
	time(&loct);
	diff = difftime(loct, svt);
	if (diff < 0) diff = -diff;
	printf("Clock diffs is %f seconds\n", diff);
	if (!doset) 
		return 0;
	if (diff > maxdiff) {
		time(&loct);
		error = ncp_SetFileServerDateAndTime(cH, &loct);
		if (error) {
			fprintf(stderr, "Can't set file server time: %04x\n", error);
			return error;
		}
		printf("Adjusted\n");
	} else {
		printf("Not touched.\n");
	}
	return 0;
}

static int
fs_exec(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;

	if (optind >= argc) {
		fprintf(stderr, "File name expected\n");
		return 1;
	}
	error = NWSMExecuteNCFFile(cH, argv[optind]);
	if (error) {
		fprintf(stderr, "Can't execute %s: %04x\n", argv[optind], error);
		return error;
	}
	return 0;
}

static int
fs_down(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error, opt, force = 0;

	while ((opt = getopt(argc, argv, "f")) != EOF) {
		switch (opt) {
		    case 'f':
			force = 1;
			break;
		    default:
			return 1;
		}
	}
	error = NWDownFileServer(cH, force);
	if (error) {
		fprintf(stderr, "Can't bring file server down: %04x\n", error);
		return error;
	}
	return 0;
}

static int
fs_openbindery(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;
	
	error = NWOpenBindery(cH);
	if (error) {
		fprintf(stderr, "Can't open bindery: %04x\n", error);
	}
	return error;
}

static int
fs_closebindery(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;

	rqpermanent();
	error = NWCloseBindery(cH);
	if (error) {
		fprintf(stderr, "Can't open bindery: %04x\n", error);
	}
	return error;
}

static int
fs_enablelogin(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;

	error = NWEnableFileServerLogin(cH);
	if (error) {
		fprintf(stderr, "Can't enable logins: %04x\n", error);
	}
	return error;
}

static int
fs_disablelogin(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;

	rqpermanent();
	error = NWDisableFileServerLogin(cH);
	if (error) {
		fprintf(stderr, "Can't disable logins: %04x\n", error);
	}
	return error;
}

static int
fs_enabletts(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;
	
	error = NWEnableTTS(cH);
	if (error) {
		fprintf(stderr, "Can't enable TTS: %04x\n", error);
	}
	return error;
}

static int
fs_disabletts(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;

	rqpermanent();
	error = NWDisableTTS(cH);
	if (error) {
		fprintf(stderr, "Can't disable TTS: %04x\n", error);
	}
	return error;
}

static int
fs_loadnlm(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;

	if (optind >= argc) {
		fprintf(stderr, "File name expected\n");
		return 1;
	}
	error = NWSMLoadNLM(cH, argv[optind]);
	if (error) {
		fprintf(stderr, "Can't load %s: %04x\n", argv[optind], error);
		return error;
	}
	return 0;
}

static int
fs_unloadnlm(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;

	if (optind >= argc) {
		fprintf(stderr, "File name expected\n");
		return 1;
	}
	error = NWSMUnloadNLM(cH, argv[optind]);
	if (error) {
		fprintf(stderr, "Can't unload %s: %04x\n", argv[optind], error);
		return error;
	}
	return 0;
}

static int
fs_mount(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;
	nuint32	volnum;

	if (optind >= argc) {
		fprintf(stderr, "Volume name expected\n");
		return 1;
	}
	error = NWSMMountVolume(cH, argv[optind], &volnum);
	if (error) {
		fprintf(stderr, "Can't mount volume %s: %04x\n", argv[optind], error);
		return error;
	}
	return 0;
}

static int
fs_dismount(NWCONN_HANDLE cH, int argc, char *argv[]) {
	int error;

	if (optind >= argc) {
		fprintf(stderr, "Volume name expected\n");
		return 1;
	}
	error = NWSMDismountVolumeByName(cH, argv[optind]);
	if (error) {
		fprintf(stderr, "Can't dismount volume %s: %04x\n", argv[optind], error);
		return error;
	}
	return 0;
}

static void
help(void) {
	printf("\n");
	printf("usage: %s [connection options] command [options]\n\n", __progname);
	printf(
	    "for connection options see ncplogin(1)\n"
	    "\n"
	    "Available commands:\n"
	    "time\n"
	    "sync [-d] [-e]\n"
	    "exec file.ncf\n"
	    "load module\n"
	    "unload module\n"
	    "mount volume\n"
	    "dismount volume\n"
	    "down [force]\n"
	    "open bindery\n"
	    "close bindery\n"
	    "enable login\n"
	    "disable login\n"
	    "enable tts\n"
	    "disable tts\n"
	    "set variable=value\n\n"
	);
}
