/*
 *
 * Copyright (C) 1999 by Boris Popov
 *
 * $Id: ncpurge.c,v 1.1.2.6 1999/09/30 10:31:02 bp Exp $
 */
#include <sys/param.h>
#include <sys/queue.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#include <netncp/ncp_lib.h>
#include <netncp/nwerror.h>
#include <netncp/ncp_file.h>

struct pf_info {
	SLIST_ENTRY(pf_info)	pf_next;
	nuint32			pf_volNumber;
	nuint32			pf_dirBase;
	nuint32			pf_seq;
	time_t			pf_deltime;
	char			pf_name[256];
};

SLIST_HEAD(pf_info_head, pf_info);


int recurse = 0;
int show = 1;
int purge = 1;
int daysold = 0;	/* by default purge all */
time_t now;
int files = 0;		/* total purgable files */
int purged = 0;		/* was successfully purged */
int failures = 0;	/* errors while purge */
static int ns = NW_NS_DOS;

static void processdir(NWCONN_HANDLE cH, char *path);
static void purgefiles(NWCONN_HANDLE cH, nuint32 pvolNumber, nuint32 pdirBase, char *path);
static int  purgable(struct pf_info *pfp);

void usage(void) {
	printf(
	    "usage: nwpurge [options] [directory]\n\n"
	    "-h         Print this help text\n"
	    "-r         Recurse into subdirectories\n"
	    "-d age     Purge only files that was deleted 'age' days ago\n"
	    "-l         List mode, do not actual purge\n"
	    "-n         Use name space from mount point instead of DOS\n"
	    "-s         Suppress file list\n"
	    "\n"
	    "directory  Directory to purge\n\n"
	);
}

int main(int argc, char* argv[]) {
	char* path, *tmp;
	NWCONN_HANDLE connHandle;
	int error, c;

	if(ncp_initlib()) exit(1);
	while ((c = getopt(argc, argv, "ad:hlnrs")) != -1) {
		switch (c) {
		    case 'r':
			recurse = 1;
			break;
		    case 'd':
			daysold = strtol(optarg, &tmp, 0);
			if (*tmp) {
				fprintf(stderr, "Invalid 'days' value: %s\n", optarg);
				exit(1);
			}
			break;
		    case 'h':
			usage(); 
			exit(2);
		    case 'l':
			purge = 0;
			break;
		    case 'n':
			ns = 0xfeff;
			break;
		    case 's':
			show = 0;
			break;
		    default: 
			exit(1);
		}
	}
	if (optind < argc) {
		path = argv[optind++];
	} else {
		path = ".";
	}
	now = time(NULL);
	error = ncp_path2conn(path, &connHandle);
	if (error) {
		fprintf(stderr, "%s is not an nwfs mount: %s\n", path, strerror(error));
		exit(1);
	}
	processdir(connHandle, path);
	ncp_disconnect(connHandle);
	printf("Purged: %d\n", purged);
	printf("Failed: %d\n", failures);
	printf(" Total: %d\n", files);
	return 0;
}

void
processdir(NWCONN_HANDLE cH, char *apath) {
	struct nw_entry_info dir;	
	NWDIR_HANDLE dh;
	char path[MAXPATHLEN*2];
	int error, entryns;

	realpath(apath, path);
	if (show)
		printf("Starting with: %s\n", path);
	error = ncp_AllocTempDirHandle(path, &dh);
	if (error) {
		fprintf(stderr, "Can't obtain dirhandle: %s\n",strerror(error));
		return;
	}
	error = ncp_GetNSEntryInfo(dh, &dir, &entryns);
	if (error) {
		fprintf(stderr, "Can't get entry info: %s\n",strerror(error));
		return;
	}
	ncp_DeallocateDirHandle(dh);
	if (ns == 0xfeff) ns = entryns;
	if ((dir.attributes & aDIR) == 0) {
		fprintf(stderr, "%s not a directory\n", path);
		return;
	}
	purgefiles(cH, dir.volNumber, dir.dirEntNum, path);
}

int
purgable(struct pf_info *pfp) {
	if (!purge) return 0;
	if (daysold == 0) return 1;
	return ((difftime(now, pfp->pf_deltime) / 86400) >= daysold);
}

static struct pf_info_head pf_list = {NULL};

void
purgefiles(NWCONN_HANDLE cH, nuint32 pvolNumber, nuint32 pdirBase, char *path) {
	int seq, error, skiped;
	nuint32 dirBase, volNum;
	NWDELETED_INFO info;
	SEARCH_SEQUENCE sq;
	struct nw_entry_info fi;
	struct pf_info *pfp, *npfp;
	char *oldpath;

	volNum = pvolNumber;
	dirBase = pdirBase;
	seq = -1;
	while ((error = ncp_ScanForDeletedFiles(cH, &seq, &volNum, &dirBase, ns,
	    &info)) == 0) {
		files++;
		pfp = malloc(sizeof(*pfp));
		strcpy(pfp->pf_name, info.name);
		pfp->pf_volNumber = volNum;
		pfp->pf_dirBase = dirBase;
		pfp->pf_seq = seq;
		pfp->pf_deltime = ncp_UnpackDateTime(info.deletedDateAndTime);
		SLIST_INSERT_HEAD(&pf_list, pfp, pf_next);
	}
	for (pfp = pf_list.slh_first; pfp; pfp = npfp) {
		npfp = pfp->pf_next.sle_next;
		error = 0;
		skiped = !purgable(pfp);
		if (!skiped) {
			error = ncp_PurgeDeletedFile(cH, pfp->pf_seq, 
			    pfp->pf_volNumber, pfp->pf_dirBase, ns);
			if (error) {
				failures++;
			} else {
				purged++;
			}
		}
		if (show) {
			char timebuf[40];
			int l;

			printf("%s/%s ", path, pfp->pf_name);
			if (error) {
				printf("failed\n");
			} else if (skiped) {
				l = strlen(ctime_r(&pfp->pf_deltime, timebuf));
				if (l) timebuf[l-1] = 0;
				printf("skiped (%s)\n", timebuf);
			} else
				printf("purged\n");
		}
		free(pfp);
	}
	pf_list.slh_first = NULL;
	if (!recurse) return;
	sq.volNumber = pvolNumber;
	sq.dirNumber = pdirBase;
	sq.searchDirNumber = -1;
	oldpath = strchr(path, 0);
	while((error = ncp_ScanNSEntryInfo(cH, ns, SA_SUBDIR_ONLY, &sq,
	    "\xff*", IM_NAME | IM_DIRECTORY, &fi)) == 0) {
		oldpath[0] = '/';
		strcpy(oldpath + 1, fi.entryName);
		purgefiles(cH, fi.volNumber, fi.DosDirNum, path);
		*oldpath = 0;
	}
}

